/*---------------------------------------------------------------------------
 
  FILENAME:
        wvutils.c
 
  PURPOSE:
        Provide some global utility API methods.
 
  REVISION HISTORY:
        Date            Engineer        Revision        Remarks
        05/29/2005      M.S. Teel       0               Original
 
  NOTES:
        
 
  LICENSE:
        Copyright (c) 2004, Mark S. Teel (mark@teel.ws)
  
        This source code is released for free distribution under the terms 
        of the GNU General Public License.
  
----------------------------------------------------------------------------*/

//  ... System header files
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <time.h>
#include <errno.h>
#include <math.h>
#include <sys/stat.h>

#include <radmsgLog.h>
#include <radbuffers.h>
#include <radconffile.h>

//  ... Local header files
#include <sysdefs.h>
#include <services.h>
#include <beaufort.h>


static USHORT VerboseMask, DaemonMask;



//  ... define methods here

// ***************************************************************************
// * Output Controlled General Event Logging
// ***************************************************************************
int wvutilsSetVerbosity (USHORT daemonBitMask)
{
    CF_ID           configFileId;
    char            instance[32], value[32];
    char            path[128];
    int             i;
    
    sprintf (path, "%s/%s", WVIEW_CONFIG_DIR, WVIEW_CONFIG_FILENAME);

    // try 4 times to get the lock
    for (i = 0; i < 4; i ++)
    {
        configFileId = radCfOpen (path);
        if (configFileId != NULL)
        {
            break;
        }
        radUtilsSleep (500);
    }
    
    // did we make it?
    if (configFileId == NULL)
    {
        return ERROR;
    }

    if (radCfGetFirstEntry (configFileId, "VERBOSE_MSGS", instance, value)
        == ERROR)
    {
        radCfClose (configFileId);
        return ERROR;
    }

    DaemonMask = daemonBitMask;

    // be backwards compatible
    if (strlen(value) < 8)
    {
        i = atoi (value);
        if (i)
            VerboseMask = WV_VERBOSE_ALL;
        else
            VerboseMask = 0x0000;
    }
    else
    {
        // process bit-by-bit
        VerboseMask = 0x0000;
        for (i = 0; i < 8; i ++)
        {
            if (value[i] == '1')
                VerboseMask |= (1 << (7-i));
        }
    }
    
    radCfClose (configFileId);
    return OK;
}

int wvutilsToggleVerbosity (void)
{
    int     retVal;

    if (DaemonMask & VerboseMask)
    {
        retVal = 0;
        VerboseMask &= ~DaemonMask;
    }
    else
    {
        retVal = 1;
        VerboseMask |= DaemonMask;
    }

    return retVal;
}

void wvutilsLogEvent (int priority, char *format, ...)
{
    va_list     argList;
    char        temp1[256];

    if ((DaemonMask & VerboseMask) == 0)
    {
        return;
    }
    else
    {
        // print the var arg stuff to the buffer
        va_start (argList, format);
        vsprintf (temp1, format, argList);
        va_end   (argList);
    
        radMsgLog (priority, temp1);
    }

    return;
}


// ***************************************************************************
// * Calculated Weather Data
// ***************************************************************************
//  calculate the heat index
float wvutilsCalculateHeatIndex (float temp, float humidity)
{
    double      T1, T2, T3, T4, T5, T6, T7, T8, T9, result;

    if (temp < 75.0)
    {
        return temp;
    }

    T1 = -42.379;
    T2 = 2.04901523 * temp;
    T3 = 10.14333127 * humidity;
    T4 = -0.22475541 * temp * humidity;
    T5 = -6.83783E-3 * (temp * temp);
    T6 = -5.481717E-2 * (humidity * humidity);
    T7 = 1.22874E-3 * (temp * temp) * humidity;
    T8 = 8.5282E-4 * temp * (humidity * humidity);
    T9 = -1.99E-6 * (temp * temp) * (humidity * humidity);
    result = T1 + T2 + T3 + T4 + T5 + T6 + T7 + T8 + T9;

    return (float)result;
}

//  calculate the wind chill
float wvutilsCalculateWindChill (float temp, float windspeed)
{
    double      T1, T2, T3, T4;
    double      result;

    if (temp >= 50.0 || windspeed <= 3.0)
    {
        return temp;
    }

    T1 = 35.74;
    T2 = 0.6215 * temp;
    T3 = -1.0 * (35.75 * pow (windspeed, 0.16));
    T4 = 0.4275 * temp * pow (windspeed, 0.16);
    result = T1 + T2 + T3 + T4;

    return (float)result;
}

//  calculate the dewpoint
float wvutilsCalculateDewpoint (float temp, float humidity)
{
    float       Tc  = (5.0/9.0)*(temp - 32.0);
    float       Es  = 6.11 * pow ((double)10.0, (double)(7.5 * (Tc/(237.7 + Tc))));
    float       E   = (humidity * Es)/100;
    float       Tdc = (-430.22 + 237.7 * log(E))/(-log(E)+19.08);
    return ((9.0/5.0) * Tdc + 32);
}

//  calculate sea level pressure from station pressure
float wvutilsCalculateSLP (float SP, float tempF, float elevationFT)
{
    double      exponent;
    double      elevMeters = (double)wvutilsConvertFeetToMeters(elevationFT);
    double      tempKelvin = (double)wvutilsConvertFahrenToCelsius(tempF) + 273.15;
    double      SLP;

    // Formula used: SLP = SP * e ^ (-mgh/RT)
    
    // molar mass of air (m) * acceleration due to gravity (g) * height (h)
    exponent = -0.0288 * 9.81 * (-elevMeters);

    // gas constant (R)
    exponent /= 8.3145;

    // degrees Kelvin (T)
    exponent /= tempKelvin;

    // compute SLP
    SLP = SP * exp(exponent);

    return (float)SLP;
}

//  calculate altimeter pressure from station pressure
float wvutilsCalculateAltimeter (float SPInches, float elevationFT)
{
    double      magicEXP            = 0.190284;
    double      inverseMagicEXP     = 1 / magicEXP;
    double      elevMeters          = (double)wvutilsConvertFeetToMeters(elevationFT);
    double      stationPressureMB   = (double)wvutilsConvertBPInchesToMillibars(SPInches);
    double      constantTerm;
    double      variableTerm;
    double      tempdouble;
    double      altimeter;

    // Formula used: http://www.wrh.noaa.gov/slc/projects/wxcalc/formulas/altimeterSetting.pdf
    
    // calculate the constant term
    constantTerm =  pow (1013.25, magicEXP);
    constantTerm *= 0.0065;
    constantTerm /= 288;

    // calculate the variable term
    tempdouble = stationPressureMB - 0.3;
    tempdouble = pow (tempdouble, magicEXP);
    variableTerm = elevMeters / tempdouble;

    // compute main term
    tempdouble = constantTerm * variableTerm;
    tempdouble += 1;
    tempdouble = pow (tempdouble, inverseMagicEXP);

    // compute altimeter
    altimeter = stationPressureMB - 0.3;
    altimeter *= tempdouble;

    // finally, convert back to inches
    altimeter *= 0.0295299;
    
    return (float)altimeter;
}

// ***************************************************************************
// * US-to-Metric Converters
// ***************************************************************************
float wvutilsConvertFahrenToCelsius (float fahrenValue)
{
    float      retVal;
    
    retVal = fahrenValue - 32.0;
    retVal *= 5.0;
    retVal /= 9.0;
    
    return retVal;
}

float wvutilsConvertDeltaFahrenToCelsius (float fahrenValue)
{
    float      retVal;
    
    retVal = fahrenValue * 5.0;
    retVal /= 9.0;
    
    return retVal;
}

float wvutilsConvertMphToKph (float mph)
{
    float      retVal;
    
    retVal = 1.609 * mph;
    
    return retVal;
}

float wvutilsConvertBPInchesToMillibars (float inches)
{
    float      retVal;
    
    retVal = inches / 0.0295299;
    
    return retVal;
}

float wvutilsConvertRainInchesToCentimeters (float inches)
{
    float      retVal;
    
    retVal = inches * 2.54;
    
    return retVal;
}

float wvutilsConvertRainInchesToMillimeters (float inches)
{
    float      retVal;
    
    retVal = inches * 25.4;
    
    return retVal;
}

float wvutilsConvertMilesToKilometers (float miles)
{
    float      retVal;
    
    retVal = miles * 1.609;
    
    return retVal;
}

float wvutilsConvertFeetToMeters (float feet)
{
    float      retVal;
    
    retVal = feet * 0.3048;
    
    return retVal;
}


//  Determine if it is day or night based on loop data
//  Returns TRUE or FALSE
int wvutilsIsDayTime (short sunrise, short sunset)
{
    time_t          currTime;
    struct tm       locTime;
    int             sunriseHour, sunriseMinute;
    int             sunsetHour, sunsetMinute;
    int             currentHour, currentMinute;
    int             isInverted = FALSE;

    // is it always day or always night?
    if (sunrise == -1)
    {
        return TRUE;
    }
    else if (sunrise == -2)
    {
        return FALSE;
    }

    // get the current time
    currTime = time (NULL);
    localtime_r (&currTime, &locTime);
    
    sunriseHour     = EXTRACT_PACKED_HOUR(sunrise);
    sunriseMinute   = EXTRACT_PACKED_MINUTE(sunrise);
    sunsetHour      = EXTRACT_PACKED_HOUR(sunset);
    sunsetMinute    = EXTRACT_PACKED_MINUTE(sunset);
    currentHour     = locTime.tm_hour;
    currentMinute   = locTime.tm_min;
    
    // determine if this is an inverted scenario (Set < Rise)
    if (sunriseHour > sunsetHour)
    {
        isInverted = TRUE;
    }
    else if (sunriseHour == sunsetHour && sunriseMinute > sunsetMinute)
    {
        isInverted = TRUE;
    }

    // walk through scenarios
    if (isInverted)
    {
        if (currentHour < sunsetHour)
        {
            return TRUE;
        }
        else if (currentHour == sunsetHour && currentMinute < sunsetMinute)
        {
            return TRUE;
        }
        else
        {
            if (currentHour < sunriseHour)
            {
                return FALSE;
            }
            else if (currentHour == sunriseHour && currentMinute < sunriseMinute)
            {
                return FALSE;
            }
            else
            {
                return TRUE;
            }
        }
    }
    else
    {
        // normal scenario
        if (currentHour < sunriseHour)
        {
            return FALSE;
        }
        else if (currentHour == sunriseHour && currentMinute < sunriseMinute)
        {
            return FALSE;
        }
        else
        {
            if (currentHour < sunsetHour)
            {
                return TRUE;
            }
            else if (currentHour == sunsetHour && currentMinute < sunsetMinute)
            {
                return TRUE;
            }
            else
            {
                return FALSE;
            }
        }
    }
}

// Air Density Calculation
static float getVaporPressure (float tempC)
{
    float       E;

    E = 6.1078 * powf (10, (7.5 * tempC / (237.3 + tempC)));

    return E;
}

static float calcDensity (float abspressmb, float e, float tc)
{
    float       Rv = 461.4964;
    float       Rd = 287.0531;
    float       tk = tc + 273.15;
    float       pv = e * 100;
    float       pd = (abspressmb - e) * 100;
    float       d = (pv/(Rv*tk)) + (pd/(Rd*tk));
    return d;
}

float wvutilsCalculateAirDensity (float tempF, float bp, float dp)
{
    float       tempC = wvutilsConvertFahrenToCelsius (tempF);
    float       bpMB = wvutilsConvertBPInchesToMillibars (bp);
    float       dewpoint = wvutilsConvertFahrenToCelsius (dp);
    float       emb = getVaporPressure (dewpoint);
    float       density = calcDensity (bpMB, emb, tempC);

    return density;
}

time_t wvutilsPackedTimeToTimeT (USHORT packedDate, USHORT packedTime)
{
    time_t      timeT;
    struct tm   locTime;

    // convert everything to epoch time
    memset (&locTime, 0, sizeof(locTime));
    locTime.tm_year = EXTRACT_PACKED_YEAR(packedDate) - 1900;
    locTime.tm_mon  = EXTRACT_PACKED_MONTH(packedDate) - 1;
    locTime.tm_mday = EXTRACT_PACKED_DAY(packedDate);
    locTime.tm_hour = EXTRACT_PACKED_HOUR(packedTime);
    locTime.tm_min  = EXTRACT_PACKED_MINUTE(packedTime);
    timeT = mktime (&locTime);
    return timeT;
}

// calculate the time difference in packed date/time format
// does not consider leap years
// returns the delta in minutes
int wvutilsCalculatePackedTimeDelta
(
    USHORT      newDate,
    USHORT      newTime,
    USHORT      oldDate,
    USHORT      oldTime
)
{
    int         diffMinutes;
    time_t      oldTimeT, newTimeT;

    // convert everything to epoch time
    oldTimeT = wvutilsPackedTimeToTimeT (oldDate, oldTime);
    newTimeT = wvutilsPackedTimeToTimeT (newDate, newTime);

    diffMinutes = (int)(newTimeT - oldTimeT);
    diffMinutes /= 60;
    return diffMinutes;
}    

// increment a packed time value by the given minutes, rolls over at 24:00
USHORT wvutilsIncrementPackedTime (USHORT pTime, int minutes)
{
    int         tempMin, tempHour;
    USHORT      retVal;

    tempMin = EXTRACT_PACKED_MINUTE(pTime) + minutes;
    tempHour = EXTRACT_PACKED_HOUR(pTime);

    if (tempMin >= 60)
    {
        tempHour += tempMin/60;
        tempMin %= 60;
    }

    if (tempHour > 24 || (tempHour == 24 && tempMin > 0))
    {
        // allow 24:00, but nothing greater
        tempHour %= 24;
    }

    retVal = (tempHour * 100) + tempMin;
    return retVal;
}
    
// produce a float string fixing the truncation annoyance
char *wvutilsPrintFloat (float value, int decPlaces)
{
    static char     flbuffer[32];
    char            format[16];
    int             intTemp;

    intTemp = (int)(value * pow (10, (decPlaces+1)));
    if ((intTemp % 10) > 4)
        intTemp += 5;
    intTemp /= 10;
    sprintf (format, "%s.%1.1d%s", "%", decPlaces, "f");
    sprintf (flbuffer, format, (float)intTemp/(pow (10, decPlaces)));
    return flbuffer;
}

int wvutilsIsDST (void)
{
    time_t              ntime;
    struct tm           tmtime;

    ntime = time (NULL);
    localtime_r (&ntime, &tmtime);
    
    return ((tmtime.tm_isdst) ? TRUE : FALSE);
}

char *wvutilsConvertToBeaufortScale (int windSpeed)
{
    static char         beaufortBfr[32];

    if (windSpeed == Beaufort_Calm)
    {
        sprintf (beaufortBfr, "Calm");
    }
    else if (windSpeed <= Beaufort_LightAir)
    {
        sprintf (beaufortBfr, "Light Air");
    }
    else if (windSpeed <= Beaufort_LightBreeze)
    {
        sprintf (beaufortBfr, "Light Breeze");
    }
    else if (windSpeed <= Beaufort_GentleBreeze)
    {
        sprintf (beaufortBfr, "Gentle Breeze");
    }
    else if (windSpeed <= Beaufort_ModerateBreeze)
    {
        sprintf (beaufortBfr, "Moderate Breeze");
    }
    else if (windSpeed <= Beaufort_FreshBreeze)
    {
        sprintf (beaufortBfr, "Fresh Breeze");
    }
    else if (windSpeed <= Beaufort_StrongBreeze)
    {
        sprintf (beaufortBfr, "Strong Breeze");
    }
    else if (windSpeed <= Beaufort_NearGale)
    {
        sprintf (beaufortBfr, "Near Gale");
    }
    else if (windSpeed <= Beaufort_Gale)
    {
        sprintf (beaufortBfr, "Gale");
    }
    else if (windSpeed <= Beaufort_SevereGale)
    {
        sprintf (beaufortBfr, "Severe Gale");
    }
    else if (windSpeed <= Beaufort_Storm)
    {
        sprintf (beaufortBfr, "Storm");
    }
    else if (windSpeed <= Beaufort_ViolentStorm)
    {
        sprintf (beaufortBfr, "Violent Storm");
    }
    else // if (windSpeed <= Beaufort_Hurricane)
    {
        sprintf (beaufortBfr, "Hurricane");
    }
   
    return beaufortBfr;
}

void wvutilsSendPMONPollResponse (int mask, PMON_PROCESS_TYPES process)
{
    WVIEW_MSG_POLL_RESPONSE     response;

    if (! PMON_PROCESS_ISSET(mask,process))
    {
        return;
    }

    response.pid = getpid ();
    radMsgRouterMessageSend (WVIEW_MSG_TYPE_POLL_RESPONSE, &response, sizeof(response));
    return;
}

